@@ -32,25 +32,38 @@ The main interface is :class:`TrinoQuery`: ::
     >> query =  TrinoQuery(request, sql)
     >> rows = list(query.execute())
 """
+from __future__ import annotations
 
+import abc
+import base64
 import copy
 import functools
 import os
 import random
 import re
 import threading
-import time
 import urllib.parse
-from datetime import datetime, timedelta, timezone
+import warnings
+from datetime import date, datetime, time, timedelta, timezone, tzinfo
 from decimal import Decimal
-from typing import Any, Dict, List, Optional, Tuple, Union
+from time import sleep
+from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union
 
 import pytz
 import requests
+from pytz.tzinfo import BaseTzInfo
+from tzlocal import get_localzone_name  # type: ignore
 
 import trino.logging
 from trino import constants, exceptions
 
+try:
+    from zoneinfo import ZoneInfo  # type: ignore
+
+except ModuleNotFoundError:
+    from backports.zoneinfo import ZoneInfo  # type: ignore
+
+
 __all__ = ["ClientSession", "TrinoQuery", "TrinoRequest", "PROXIES"]
 
 logger = trino.logging.get_logger(__name__)
@@ -64,9 +77,16 @@ else:
 
 _HEADER_EXTRA_CREDENTIAL_KEY_REGEX = re.compile(r'^\S[^\s=]*$')
 
-INF = float("inf")
-NEGATIVE_INF = float("-inf")
-NAN = float("nan")
+T = TypeVar("T")
+
+PythonTemporalType = TypeVar("PythonTemporalType", bound=Union[time, datetime])
+POWERS_OF_TEN: Dict[int, Decimal] = {}
+for i in range(0, 13):
+    POWERS_OF_TEN[i] = Decimal(10 ** i)
+MAX_PYTHON_TEMPORAL_PRECISION_POWER = 6
+MAX_PYTHON_TEMPORAL_PRECISION = POWERS_OF_TEN[MAX_PYTHON_TEMPORAL_PRECISION_POWER]
+
+ROLE_PATTERN = re.compile(r"^ROLE\{(.*)\}$")
 
 
 class ClientSession(object):
@@ -100,6 +120,7 @@ class ClientSession(object):
     :param client_tags: Client tags as list of strings.
     :param roles: roles for the current session. Some connectors do not
                  support role management. See connector documentation for more details.
+    :param timezone: The timezone for query processing. Defaults to the system's local timezone.
     """
 
     def __init__(
@@ -114,6 +135,7 @@ class ClientSession(object):
         extra_credential: List[Tuple[str, str]] = None,
         client_tags: List[str] = None,
         roles: Dict[str, str] = None,
+        timezone: str = None,
     ):
         self._user = user
         self._catalog = catalog
@@ -124,9 +146,12 @@ class ClientSession(object):
         self._transaction_id = transaction_id
         self._extra_credential = extra_credential
         self._client_tags = client_tags.copy() if client_tags is not None else list()
-        self._roles = roles.copy() if roles is not None else {}
+        self._roles = self._format_roles(roles) if roles is not None else {}
         self._prepared_statements: Dict[str, str] = {}
         self._object_lock = threading.Lock()
+        self._timezone = timezone or get_localzone_name()
+        if timezone:  # Check timezone validity
+            ZoneInfo(timezone)
 
     @property
     def user(self):
@@ -207,6 +232,26 @@ class ClientSession(object):
         with self._object_lock:
             self._prepared_statements = prepared_statements
 
+    @property
+    def timezone(self):
+        with self._object_lock:
+            return self._timezone
+
+    def _format_roles(self, roles):
+        formatted_roles = {}
+        for catalog, role in roles.items():
+            is_legacy_role_pattern = ROLE_PATTERN.match(role) is not None
+            if role in ("NONE", "ALL") or is_legacy_role_pattern:
+                if is_legacy_role_pattern:
+                    warnings.warn(f"A role '{role}' is provided using a legacy format. "
+                                  "Please remove the ROLE{} wrapping. Support for the legacy format might be "
+                                  "removed in a future release.",
+                                  DeprecationWarning)
+                formatted_roles[catalog] = role
+            else:
+                formatted_roles[catalog] = f"ROLE{{{role}}}"
+        return formatted_roles
+
     def __getstate__(self):
         state = self.__dict__.copy()
         del state["_object_lock"]
@@ -295,7 +340,7 @@ class _RetryWithExponentialBackoff(object):
 
     def retry(self, func, args, kwargs, err, attempt):
         delay = self._get_delay(attempt)
-        time.sleep(delay)
+        sleep(delay)
 
 
 class TrinoRequest(object):
@@ -408,6 +453,8 @@ class TrinoRequest(object):
         headers[constants.HEADER_SCHEMA] = self._client_session.schema
         headers[constants.HEADER_SOURCE] = self._client_session.source
         headers[constants.HEADER_USER] = self._client_session.user
+        headers[constants.HEADER_TIMEZONE] = self._client_session.timezone
+        headers[constants.HEADER_CLIENT_CAPABILITIES] = 'PARAMETRIC_DATETIME'
         if len(self._client_session.roles.values()):
             headers[constants.HEADER_ROLE] = ",".join(
                 # ``name`` must not contain ``=``
@@ -494,7 +541,7 @@ class TrinoRequest(object):
     def next_uri(self) -> Optional[str]:
         return self._next_uri
 
-    def post(self, sql, additional_http_headers=None):
+    def post(self, sql: str, additional_http_headers: Optional[Dict[str, Any]] = None):
         data = sql.encode("utf-8")
         # Deep copy of the http_headers dict since they may be modified for this
         # request by the provided additional_http_headers
@@ -526,7 +573,7 @@ class TrinoRequest(object):
                 )
         return http_response
 
-    def get(self, url):
+    def get(self, url: str):
         return self._get(
             url,
             headers=self.http_headers,
@@ -683,7 +730,7 @@ class TrinoQuery(object):
             self,
             request: TrinoRequest,
             sql: str,
-            experimental_python_types: bool = False,
+            legacy_primitive_types: bool = False,
     ) -> None:
         self.query_id: Optional[str] = None
 
@@ -697,7 +744,7 @@ class TrinoQuery(object):
         self._update_type = None
         self._sql = sql
         self._result: Optional[TrinoResult] = None
-        self._experimental_python_types = experimental_python_types
+        self._legacy_primitive_types = legacy_primitive_types
         self._row_mapper: Optional[RowMapper] = None
 
     @property
@@ -763,7 +810,7 @@ class TrinoQuery(object):
         self._update_type = status.update_type
         if not self._row_mapper and status.columns:
             self._row_mapper = RowMapperFactory().create(columns=status.columns,
-                                                         experimental_python_types=self._experimental_python_types)
+                                                         legacy_primitive_types=self._legacy_primitive_types)
         if status.columns:
             self._columns = status.columns
 
@@ -840,10 +887,257 @@ def _retry_with(handle_retry, handled_exceptions, conditions, max_attempts):
     return wrapper
 
 
+class ValueMapper(abc.ABC, Generic[T]):
+    @abc.abstractmethod
+    def map(self, value: Any) -> Optional[T]:
+        pass
+
+
+class NoOpValueMapper(ValueMapper[Any]):
+    def map(self, value) -> Optional[Any]:
+        return value
+
+
+class DecimalValueMapper(ValueMapper[Decimal]):
+    def map(self, value) -> Optional[Decimal]:
+        if value is None:
+            return None
+        return Decimal(value)
+
+
+class DoubleValueMapper(ValueMapper[float]):
+    def map(self, value) -> Optional[float]:
+        if value is None:
+            return None
+        if value == 'Infinity':
+            return float("inf")
+        if value == '-Infinity':
+            return float("-inf")
+        if value == 'NaN':
+            return float("nan")
+        return float(value)
+
+
+def _create_tzinfo(timezone_str: str) -> tzinfo:
+    if timezone_str.startswith("+") or timezone_str.startswith("-"):
+        hours = timezone_str[1:3]
+        minutes = timezone_str[4:6]
+        if timezone_str.startswith("-"):
+            return timezone(-timedelta(hours=int(hours), minutes=int(minutes)))
+        return timezone(timedelta(hours=int(hours), minutes=int(minutes)))
+    else:
+        return pytz.timezone(timezone_str)
+
+
+def _fraction_to_decimal(fractional_str: str) -> Decimal:
+    return Decimal(fractional_str or 0) / POWERS_OF_TEN[len(fractional_str)]
+
+
+class TemporalType(Generic[PythonTemporalType], metaclass=abc.ABCMeta):
+    def __init__(self, whole_python_temporal_value: PythonTemporalType, remaining_fractional_seconds: Decimal):
+        self._whole_python_temporal_value = whole_python_temporal_value
+        self._remaining_fractional_seconds = remaining_fractional_seconds
+
+    @abc.abstractmethod
+    def new_instance(self, value: PythonTemporalType, fraction: Decimal) -> TemporalType[PythonTemporalType]:
+        pass
+
+    @abc.abstractmethod
+    def to_python_type(self) -> PythonTemporalType:
+        pass
+
+    def round_to(self, precision: int) -> TemporalType:
+        """
+            Python datetime and time only support up to microsecond precision
+            In case the supplied value exceeds the specified precision,
+            the value needs to be rounded.
+        """
+        precision = min(precision, MAX_PYTHON_TEMPORAL_PRECISION_POWER)
+        remaining_fractional_seconds = self._remaining_fractional_seconds
+        digits = abs(remaining_fractional_seconds.as_tuple().exponent)
+        if digits > precision:
+            rounding_factor = POWERS_OF_TEN[precision]
+            rounded = remaining_fractional_seconds.quantize(Decimal(1 / rounding_factor))
+            if rounded == rounding_factor:
+                return self.new_instance(
+                    self.normalize(self.add_time_delta(timedelta(seconds=1))),
+                    Decimal(0)
+                )
+            return self.new_instance(self._whole_python_temporal_value, rounded)
+        return self
+
+    @abc.abstractmethod
+    def add_time_delta(self, time_delta: timedelta) -> PythonTemporalType:
+        """
+            This method shall be overriden to implement fraction arithmetics.
+        """
+        pass
+
+    def normalize(self, value: PythonTemporalType) -> PythonTemporalType:
+        """
+            If `add_time_delta` results in value crossing DST boundaries, this method should
+            return a normalized version of the value to account for it, for example,
+            using `pytz.timezone.normalize`.
+        """
+        return value
+
+
+class Time(TemporalType[time]):
+    def new_instance(self, value: time, fraction: Decimal) -> TemporalType[time]:
+        return Time(value, fraction)
+
+    def to_python_type(self) -> time:
+        if self._remaining_fractional_seconds > 0:
+            time_delta = timedelta(microseconds=int(self._remaining_fractional_seconds * MAX_PYTHON_TEMPORAL_PRECISION))
+            return self.add_time_delta(time_delta)
+        return self._whole_python_temporal_value
+
+    def add_time_delta(self, time_delta: timedelta) -> time:
+        time_delta_added = datetime.combine(datetime(1, 1, 1), self._whole_python_temporal_value) + time_delta
+        return time_delta_added.time().replace(tzinfo=self._whole_python_temporal_value.tzinfo)
+
+
+class TimeWithTimeZone(Time, TemporalType[time]):
+    def new_instance(self, value: time, fraction: Decimal) -> TemporalType[time]:
+        return TimeWithTimeZone(value, fraction)
+
+
+class Timestamp(TemporalType[datetime]):
+    def new_instance(self, value: datetime, fraction: Decimal) -> Timestamp:
+        return Timestamp(value, fraction)
+
+    def to_python_type(self) -> datetime:
+        if self._remaining_fractional_seconds > 0:
+            time_delta = timedelta(microseconds=int(self._remaining_fractional_seconds * MAX_PYTHON_TEMPORAL_PRECISION))
+            return self.add_time_delta(time_delta)
+        return self._whole_python_temporal_value
+
+    def add_time_delta(self, time_delta: timedelta) -> datetime:
+        return self._whole_python_temporal_value + time_delta
+
+
+class TimestampWithTimeZone(Timestamp, TemporalType[datetime]):
+    def new_instance(self, value: datetime, fraction: Decimal) -> TimestampWithTimeZone:
+        return TimestampWithTimeZone(value, fraction)
+
+    def normalize(self, value: datetime) -> datetime:
+        if isinstance(self._whole_python_temporal_value.tzinfo, BaseTzInfo):
+            return self._whole_python_temporal_value.tzinfo.normalize(value)
+        return value
+
+
+class TimeValueMapper(ValueMapper[time]):
+    def __init__(self, precision):
+        self.time_default_size = 8  # size of 'HH:MM:SS'
+        self.precision = precision
+
+    def map(self, value) -> Optional[time]:
+        if value is None:
+            return None
+        whole_python_temporal_value = value[:self.time_default_size]
+        remaining_fractional_seconds = value[self.time_default_size + 1:]
+        return Time(
+            time.fromisoformat(whole_python_temporal_value),
+            _fraction_to_decimal(remaining_fractional_seconds)
+        ).round_to(self.precision).to_python_type()
+
+    def _add_second(self, time_value: time) -> time:
+        return (datetime.combine(datetime(1, 1, 1), time_value) + timedelta(seconds=1)).time()
+
+
+class TimeWithTimeZoneValueMapper(TimeValueMapper):
+    def map(self, value) -> Optional[time]:
+        if value is None:
+            return None
+        whole_python_temporal_value = value[:self.time_default_size]
+        remaining_fractional_seconds = value[self.time_default_size + 1:len(value) - 6]
+        timezone_part = value[len(value) - 6:]
+        return TimeWithTimeZone(
+            time.fromisoformat(whole_python_temporal_value).replace(tzinfo=_create_tzinfo(timezone_part)),
+            _fraction_to_decimal(remaining_fractional_seconds),
+        ).round_to(self.precision).to_python_type()
+
+
+class DateValueMapper(ValueMapper[date]):
+    def map(self, value) -> Optional[date]:
+        if value is None:
+            return None
+        return date.fromisoformat(value)
+
+
+class TimestampValueMapper(ValueMapper[datetime]):
+    def __init__(self, precision):
+        self.datetime_default_size = 19  # size of 'YYYY-MM-DD HH:MM:SS' (the datetime string up to the seconds)
+        self.precision = precision
+
+    def map(self, value) -> Optional[datetime]:
+        if value is None:
+            return None
+        whole_python_temporal_value = value[:self.datetime_default_size]
+        remaining_fractional_seconds = value[self.datetime_default_size + 1:]
+        return Timestamp(
+            datetime.fromisoformat(whole_python_temporal_value),
+            _fraction_to_decimal(remaining_fractional_seconds),
+        ).round_to(self.precision).to_python_type()
+
+
+class TimestampWithTimeZoneValueMapper(TimestampValueMapper):
+    def map(self, value) -> Optional[datetime]:
+        if value is None:
+            return None
+        datetime_with_fraction, timezone_part = value.rsplit(' ', 1)
+        whole_python_temporal_value = datetime_with_fraction[:self.datetime_default_size]
+        remaining_fractional_seconds = datetime_with_fraction[self.datetime_default_size + 1:]
+        return TimestampWithTimeZone(
+            datetime.fromisoformat(whole_python_temporal_value).replace(tzinfo=_create_tzinfo(timezone_part)),
+            _fraction_to_decimal(remaining_fractional_seconds),
+        ).round_to(self.precision).to_python_type()
+
+
+class BinaryValueMapper(ValueMapper[bytes]):
+    def map(self, value) -> Optional[bytes]:
+        if value is None:
+            return None
+        return base64.b64decode(value.encode("utf8"))
+
+
+class ArrayValueMapper(ValueMapper[List[Optional[Any]]]):
+    def __init__(self, mapper: ValueMapper[Any]):
+        self.mapper = mapper
+
+    def map(self, values: List[Any]) -> Optional[List[Any]]:
+        if values is None:
+            return None
+        return [self.mapper.map(value) for value in values]
+
+
+class RowValueMapper(ValueMapper[Tuple[Optional[Any], ...]]):
+    def __init__(self, mappers: List[ValueMapper[Any]]):
+        self.mappers = mappers
+
+    def map(self, values: List[Any]) -> Optional[Tuple[Optional[Any], ...]]:
+        if values is None:
+            return None
+        return tuple(self.mappers[index].map(value) for index, value in enumerate(values))
+
+
+class MapValueMapper(ValueMapper[Dict[Any, Optional[Any]]]):
+    def __init__(self, key_mapper: ValueMapper[Any], value_mapper: ValueMapper[Any]):
+        self.key_mapper = key_mapper
+        self.value_mapper = value_mapper
+
+    def map(self, values: Any) -> Optional[Dict[Any, Optional[Any]]]:
+        if values is None:
+            return None
+        return {
+            self.key_mapper.map(key): self.value_mapper.map(value) for key, value in values.items()
+        }
+
+
 class NoOpRowMapper:
     """
     No-op RowMapper which does not perform any transformation
-    Used when experimental_python_types is False.
+    Used when legacy_primitive_types is False.
     """
 
     def map(self, rows):
@@ -856,117 +1150,52 @@ class RowMapperFactory:
     lambda functions (one for each column) which will process a data value
     and returns a RowMapper instance which will process rows of data
     """
-    no_op_row_mapper = NoOpRowMapper()
+    NO_OP_ROW_MAPPER = NoOpRowMapper()
 
-    def create(self, columns, experimental_python_types):
+    def create(self, columns, legacy_primitive_types):
         assert columns is not None
 
-        if experimental_python_types:
-            return RowMapper([self._col_func(column['typeSignature']) for column in columns])
-        return RowMapperFactory.no_op_row_mapper
+        if not legacy_primitive_types:
+            return RowMapper([self._create_value_mapper(column['typeSignature']) for column in columns])
+        return RowMapperFactory.NO_OP_ROW_MAPPER
 
-    def _col_func(self, column):
+    def _create_value_mapper(self, column) -> ValueMapper:
         col_type = column['rawType']
 
         if col_type == 'array':
-            return self._array_map_func(column)
+            value_mapper = self._create_value_mapper(column['arguments'][0]['value'])
+            return ArrayValueMapper(value_mapper)
         elif col_type == 'row':
-            return self._row_map_func(column)
+            mappers = [self._create_value_mapper(arg['value']['typeSignature']) for arg in column['arguments']]
+            return RowValueMapper(mappers)
         elif col_type == 'map':
-            return self._map_map_func(column)
+            key_mapper = self._create_value_mapper(column['arguments'][0]['value'])
+            value_mapper = self._create_value_mapper(column['arguments'][1]['value'])
+            return MapValueMapper(key_mapper, value_mapper)
         elif col_type.startswith('decimal'):
-            return lambda val: Decimal(val)
+            return DecimalValueMapper()
         elif col_type.startswith('double') or col_type.startswith('real'):
-            return self._double_map_func()
+            return DoubleValueMapper()
+        elif col_type.startswith('timestamp') and 'with time zone' in col_type:
+            return TimestampWithTimeZoneValueMapper(self._get_precision(column))
         elif col_type.startswith('timestamp'):
-            return self._timestamp_map_func(column, col_type)
+            return TimestampValueMapper(self._get_precision(column))
+        elif col_type.startswith('time') and 'with time zone' in col_type:
+            return TimeWithTimeZoneValueMapper(self._get_precision(column))
         elif col_type.startswith('time'):
-            return self._time_map_func(column, col_type)
+            return TimeValueMapper(self._get_precision(column))
         elif col_type == 'date':
-            return lambda val: datetime.strptime(val, '%Y-%m-%d').date()
+            return DateValueMapper()
+        elif col_type == 'varbinary':
+            return BinaryValueMapper()
         else:
-            return lambda val: val
+            return NoOpValueMapper()
 
-    def _array_map_func(self, column):
-        element_mapping_func = self._col_func(column['arguments'][0]['value'])
-        return lambda values: [element_mapping_func(value) for value in values]
-
-    def _row_map_func(self, column):
-        element_mapping_func = [self._col_func(arg['value']['typeSignature']) for arg in column['arguments']]
-        return lambda values: tuple(element_mapping_func[idx](value) for idx, value in enumerate(values))
-
-    def _map_map_func(self, column):
-        key_mapping_func = self._col_func(column['arguments'][0]['value'])
-        value_mapping_func = self._col_func(column['arguments'][1]['value'])
-        return lambda values: {key_mapping_func(key): value_mapping_func(value) for key, value in values.items()}
-
-    def _double_map_func(self):
-        return lambda val: INF if val == 'Infinity' \
-            else NEGATIVE_INF if val == '-Infinity' \
-            else NAN if val == 'NaN' \
-            else float(val)
-
-    def _timestamp_map_func(self, column, col_type):
-        datetime_default_size = 20  # size of 'YYYY-MM-DD HH:MM:SS.' (the datetime string up to the milliseconds)
-        pattern = "%Y-%m-%d %H:%M:%S"
-        ms_size, ms_to_trim = self._get_number_of_digits(column)
-        if ms_size > 0:
-            pattern += ".%f"
-
-        dt_size = datetime_default_size + ms_size - ms_to_trim
-        dt_tz_offset = datetime_default_size + ms_size
-        if 'with time zone' in col_type:
-
-            if ms_to_trim > 0:
-                return lambda val: \
-                    [datetime.strptime(val[:dt_size] + val[dt_tz_offset:], pattern + ' %z')
-                        if tz.startswith('+') or tz.startswith('-')
-                        else datetime.strptime(dt[:dt_size] + dt[dt_tz_offset:], pattern)
-                                     .replace(tzinfo=pytz.timezone(tz))
-                        for dt, tz in [val.rsplit(' ', 1)]][0]
-            else:
-                return lambda val: [datetime.strptime(val, pattern + ' %z')
-                                    if tz.startswith('+') or tz.startswith('-')
-                                    else datetime.strptime(dt, pattern).replace(tzinfo=pytz.timezone(tz))
-                                    for dt, tz in [val.rsplit(' ', 1)]][0]
-
-        if ms_to_trim > 0:
-            return lambda val: datetime.strptime(val[:dt_size] + val[dt_tz_offset:], pattern)
-        else:
-            return lambda val: datetime.strptime(val, pattern)
-
-    def _time_map_func(self, column, col_type):
-        pattern = "%H:%M:%S"
-        ms_size, ms_to_trim = self._get_number_of_digits(column)
-        if ms_size > 0:
-            pattern += ".%f"
-
-        time_size = 9 + ms_size - ms_to_trim
-
-        if 'with time zone' in col_type:
-            return lambda val: self._get_time_with_timezome(val, time_size, pattern)
-        else:
-            return lambda val: datetime.strptime(val[:time_size], pattern).time()
-
-    def _get_time_with_timezome(self, value, time_size, pattern):
-        matches = re.match(r'^(.*)([\+\-])(\d{2}):(\d{2})$', value)
-        assert matches is not None
-        assert len(matches.groups()) == 4
-        if matches.group(2) == '-':
-            tz = -timedelta(hours=int(matches.group(3)), minutes=int(matches.group(4)))
-        else:
-            tz = timedelta(hours=int(matches.group(3)), minutes=int(matches.group(4)))
-        return datetime.strptime(matches.group(1)[:time_size], pattern).time().replace(tzinfo=timezone(tz))
-
-    def _get_number_of_digits(self, column):
+    def _get_precision(self, column: Dict[str, Any]):
         args = column['arguments']
         if len(args) == 0:
-            return 3, 0
-        ms_size = column['arguments'][0]['value']
-        if ms_size == 0:
-            return -1, 0
-        ms_to_trim = ms_size - min(ms_size, 6)
-        return ms_size, ms_to_trim
+            return 3
+        return args[0]['value']
 
 
 class RowMapper:
@@ -982,14 +1211,11 @@ class RowMapper:
         return [self._map_row(row) for row in rows]
 
     def _map_row(self, row):
-        return [self._map_value(value, self.columns[idx]) for idx, value in enumerate(row)]
-
-    def _map_value(self, value, col_mapping_func):
-        if value is None:
-            return None
+        return [self._map_value(value, self.columns[index]) for index, value in enumerate(row)]
 
+    def _map_value(self, value, value_mapper: ValueMapper[T]) -> Optional[T]:
         try:
-            return col_mapping_func(value)
+            return value_mapper.map(value)
         except ValueError as e:
             error_str = f"Could not convert '{value}' into the associated python type"
             raise trino.exceptions.TrinoDataError(error_str) from e
