@@ -17,16 +17,18 @@ https://www.python.org/dev/peps/pep-0249/ .
 Fetch methods returns rows as a list of lists on purpose to let the caller
 decide to convert then to a list of tuples.
 """
+import binascii
 import datetime
 import math
 import uuid
 from decimal import Decimal
-from typing import Any, List, Optional  # NOQA for mypy types
+from typing import Any, Dict, List, NamedTuple, Optional  # NOQA for mypy types
 
 import trino.client
 import trino.exceptions
 import trino.logging
 from trino import constants
+from trino.constants import LENGTH_TYPES, PRECISION_TYPES, SCALE_TYPES
 from trino.exceptions import (
     DatabaseError,
     DataError,
@@ -108,8 +110,9 @@ class Connection(object):
         verify=True,
         http_session=None,
         client_tags=None,
-        experimental_python_types=False,
+        legacy_primitive_types=False,
         roles=None,
+        timezone=None,
     ):
         self.host = host
         self.port = port
@@ -129,6 +132,7 @@ class Connection(object):
             extra_credential=extra_credential,
             client_tags=client_tags,
             roles=roles,
+            timezone=timezone,
         )
         # mypy cannot follow module import
         if http_session is None:
@@ -148,7 +152,7 @@ class Connection(object):
         self._isolation_level = isolation_level
         self._request = None
         self._transaction = None
-        self.experimental_python_types = experimental_python_types
+        self.legacy_primitive_types = legacy_primitive_types
 
     @property
     def isolation_level(self):
@@ -170,9 +174,8 @@ class Connection(object):
             self.close()
 
     def close(self):
-        """Trino does not have anything to close"""
         # TODO cancel outstanding queries?
-        pass
+        self._http_session.close()
 
     def start_transaction(self):
         self._transaction = Transaction(self._create_request())
@@ -204,7 +207,7 @@ class Connection(object):
             self.request_timeout,
         )
 
-    def cursor(self, experimental_python_types: bool = None):
+    def cursor(self, legacy_primitive_types: bool = None):
         """Return a new :py:class:`Cursor` object using the connection."""
         if self.isolation_level != IsolationLevel.AUTOCOMMIT:
             if self.transaction is None:
@@ -216,8 +219,47 @@ class Connection(object):
         return Cursor(
             self,
             request,
-            # if experimental_python_types is not explicitly set in Cursor, take from Connection
-            experimental_python_types if experimental_python_types is not None else self.experimental_python_types
+            # if legacy_primitive_types is not explicitly set in Cursor, take from Connection
+            legacy_primitive_types if legacy_primitive_types is not None else self.legacy_primitive_types
+        )
+
+
+class DescribeOutput(NamedTuple):
+    name: str
+    catalog: str
+    schema: str
+    table: str
+    type: str
+    type_size: int
+    aliased: bool
+
+    @classmethod
+    def from_row(cls, row: List[Any]):
+        return cls(*row)
+
+
+class ColumnDescription(NamedTuple):
+    name: str
+    type_code: int
+    display_size: int
+    internal_size: int
+    precision: int
+    scale: int
+    null_ok: bool
+
+    @classmethod
+    def from_column(cls, column: Dict[str, Any]):
+        type_signature = column["typeSignature"]
+        raw_type = type_signature["rawType"]
+        arguments = type_signature["arguments"]
+        return cls(
+            column["name"],  # name
+            column["type"],  # type_code
+            None,  # display_size
+            arguments[0]["value"] if raw_type in LENGTH_TYPES else None,  # internal_size
+            arguments[0]["value"] if raw_type in PRECISION_TYPES else None,  # precision
+            arguments[1]["value"] if raw_type in SCALE_TYPES else None,  # scale
+            None  # null_ok
         )
 
 
@@ -229,7 +271,7 @@ class Cursor(object):
 
     """
 
-    def __init__(self, connection, request, experimental_python_types: bool = False):
+    def __init__(self, connection, request, legacy_primitive_types: bool = False):
         if not isinstance(connection, Connection):
             raise ValueError(
                 "connection must be a Connection object: {}".format(type(connection))
@@ -240,7 +282,7 @@ class Cursor(object):
         self.arraysize = 1
         self._iterator = None
         self._query = None
-        self._experimental_pyton_types = experimental_python_types
+        self._legacy_primitive_types = legacy_primitive_types
 
     def __iter__(self):
         return self._iterator
@@ -262,14 +304,13 @@ class Cursor(object):
         return None
 
     @property
-    def description(self):
+    def description(self) -> List[ColumnDescription]:
         if self._query.columns is None:
             return None
 
         # [ (name, type_code, display_size, internal_size, precision, scale, null_ok) ]
         return [
-            (col["name"], col["type"], None, None, None, None, None)
-            for col in self._query.columns
+            ColumnDescription.from_column(col) for col in self._query.columns
         ]
 
     @property
@@ -289,6 +330,12 @@ class Cursor(object):
             return self._query.stats
         return None
 
+    @property
+    def query_id(self) -> Optional[str]:
+        if self._query is not None:
+            return self._query.query_id
+        return None
+
     @property
     def warnings(self):
         if self._query is not None:
@@ -311,7 +358,7 @@ class Cursor(object):
         """
         sql = f"PREPARE {name} FROM {statement}"
         query = trino.client.TrinoQuery(self.connection._create_request(), sql=sql,
-                                        experimental_python_types=self._experimental_pyton_types)
+                                        legacy_primitive_types=self._legacy_primitive_types)
         query.execute()
 
     def _execute_prepared_statement(
@@ -320,7 +367,7 @@ class Cursor(object):
         params
     ):
         sql = 'EXECUTE ' + statement_name + ' USING ' + ','.join(map(self._format_prepared_param, params))
-        return trino.client.TrinoQuery(self._request, sql=sql, experimental_python_types=self._experimental_pyton_types)
+        return trino.client.TrinoQuery(self._request, sql=sql, legacy_primitive_types=self._legacy_primitive_types)
 
     def _format_prepared_param(self, param):
         """
@@ -393,12 +440,15 @@ class Cursor(object):
         if isinstance(param, Decimal):
             return "DECIMAL '%s'" % param
 
+        if isinstance(param, (bytes, bytearray)):
+            return "X'%s'" % binascii.hexlify(param).decode("utf-8")
+
         raise trino.exceptions.NotSupportedError("Query parameter of type '%s' is not supported." % type(param))
 
     def _deallocate_prepared_statement(self, statement_name: str) -> None:
         sql = 'DEALLOCATE PREPARE ' + statement_name
         query = trino.client.TrinoQuery(self.connection._create_request(), sql=sql,
-                                        experimental_python_types=self._experimental_pyton_types)
+                                        legacy_primitive_types=self._legacy_primitive_types)
         query.execute()
 
     def _generate_unique_statement_name(self):
@@ -420,7 +470,7 @@ class Cursor(object):
                 self._query = self._execute_prepared_statement(
                     statement_name, params
                 )
-                result = self._query.execute()
+                self._iterator = iter(self._query.execute())
             finally:
                 # Send deallocate statement
                 # At this point the query can be deallocated since it has already
@@ -430,10 +480,9 @@ class Cursor(object):
 
         else:
             self._query = trino.client.TrinoQuery(self._request, sql=operation,
-                                                  experimental_python_types=self._experimental_pyton_types)
-            result = self._query.execute()
-        self._iterator = iter(result)
-        return result
+                                                  legacy_primitive_types=self._legacy_primitive_types)
+            self._iterator = iter(self._query.execute())
+        return self
 
     def executemany(self, operation, seq_of_params):
         """
@@ -460,6 +509,7 @@ class Cursor(object):
             self.execute(operation, seq_of_params[-1])
         else:
             self.execute(operation)
+        return self
 
     def fetchone(self) -> Optional[List[Any]]:
         """
@@ -512,6 +562,28 @@ class Cursor(object):
 
         return result
 
+    def describe(self, sql: str) -> List[DescribeOutput]:
+        """
+        List the output columns of a SQL statement, including the column name (or alias), catalog, schema, table, type,
+        type size in bytes, and a boolean indicating if the column is aliased.
+
+        :param sql: SQL statement
+        """
+        statement_name = self._generate_unique_statement_name()
+        self._prepare_statement(sql, statement_name)
+        try:
+            sql = f"DESCRIBE OUTPUT {statement_name}"
+            self._query = trino.client.TrinoQuery(
+                self._request,
+                sql=sql,
+                legacy_primitive_types=self._legacy_primitive_types,
+            )
+            result = self._query.execute()
+        finally:
+            self._deallocate_prepared_statement(statement_name)
+
+        return list(map(lambda x: DescribeOutput.from_row(x), result))
+
     def genall(self):
         return self._query.result
 
